class Textbox { // 1-line simple textbox. ENTER will issue the typed data to the program and take out the focus
  PImage trickery; // background image to hide old text
  PGraphics textbox, border;

  int Border;
  int TWidth, THeight;
  int x, y;

  boolean OnFocus;

  String typed, cursorbuffer;    // text buffers
  int CursorCharIndex;
  int Fontsize;
  color TColour;
  float Margin; // in percentage (safer)
 
  Textbox(int _x, int _y, int _TWidth, int _THeight, String _defaulttext, int _Border, float _Margin, color _TColour) {
    x = _x; 
    y = _y; 
    TWidth = _TWidth; 
    THeight = _THeight;
    typed = cursorbuffer = _defaulttext;

    Border = _Border;

    textbox = createGraphics(TWidth - 2*Border, THeight - 2*Border, P2D);  // textbox area
    trickery = createImage(TWidth, THeight, RGB);    // background image

    border = createGraphics(TWidth, THeight, P2D);   // border
    OnFocus = false;
    CursorCharIndex = typed.length();
    CalculateFont();
    textbox.imageMode(CENTER);
    border.imageMode(CENTER);
    TColour = _TColour;
    Margin = _Margin;
    
  }

  void CalculateFont() {
    Fontsize = THeight - 2*Border;
  }

  void buildtrickery() {
    int i, px, py;
    color TMPColour, GETColour;

    trickery.loadPixels();
    for (i = 0; i < trickery.pixels.length; i++) {
      px = i%TWidth;
      py = i/TWidth;
      GETColour = menuL.get(x -int(TWidth/2) + (i % TWidth), y -int(THeight/2) + (i/TWidth));
      if ( (py < Border && py < px) || (px >= TWidth - Border && py <= px - (TWidth - THeight)) ) {
        TMPColour = lerpColor(color(0), GETColour, 0.8);
      }
      else if (( py >= THeight - Border ) || ( px < Border )) {
        TMPColour = lerpColor(color(0), GETColour, 0.8);
      }
      else TMPColour = GETColour;
      
      trickery.pixels[i] = color(0);
    }
    trickery.updatePixels();
  }

  void KeyControl() {
    if (key == BACKSPACE && CursorCharIndex > 0) {
      typed = typed.substring(0, CursorCharIndex-1) + typed.substring(CursorCharIndex, typed.length());
      //cursorbuffer = typed.substring(0, CursorCharIndex-1);
      CursorCharIndex--;
      //cursorbuffer = typed.substring(0, CursorCharIndex);
    }
    else if (key == DELETE && CursorCharIndex < typed.length()) {
      typed = typed.substring(0, CursorCharIndex) + typed.substring(CursorCharIndex+1, typed.length());
      //cursorbuffer = typed.substring(0, CursorCharIndex);
    }
    else if (key != CODED && (key != ENTER && key != RETURN) && key != DELETE && key != BACKSPACE) {
      typed = typed.substring(0, CursorCharIndex) + key + typed.substring(CursorCharIndex, typed.length());
      CursorCharIndex++;
      //cursorbuffer = typed.substring(0, CursorCharIndex);
    }
    if (key == ENTER || key == RETURN) OnFocus = false;

    if (key == CODED) {
      switch(keyCode) {
      case LEFT:
        if (CursorCharIndex > 0) { 
          CursorCharIndex--; 
        }//cursorbuffer = typed.substring(0, CursorCharIndex); }
        break;
      case RIGHT:
        if (CursorCharIndex < typed.length()) { 
          CursorCharIndex++; 
        }//cursorbuffer = typed.substring(0, CursorCharIndex); }
        break;
      case KeyEvent.VK_END:
        CursorCharIndex = typed.length(); //cursorbuffer = typed.substring(0, CursorCharIndex);
        break;
      case KeyEvent.VK_HOME:
        CursorCharIndex = 0; //cursorbuffer = typed.substring(0, CursorCharIndex);
        break;
      }
    }
    cursorbuffer = typed.substring(0, CursorCharIndex);
  }

  void keyPressed() {
    if (OnFocus) KeyControl();
    
  }

  void mouseClicked() {
    float dx, dy;
    int i, besti;
    float prevcharpos, nextcharpos;

    dx = abs(mouseX - x); // > 0: to the right of center ; < 0: to the left of center
    dy = abs(mouseY - y);

    if (dx <= TWidth/2 && dy <= THeight/2) {
      OnFocus = true;
      if (typed.length() > 0) {
        prevcharpos = x + 5 - textbox.textWidth(typed)/2; 
        besti = -1;  // that +5 comes from the fact that ControlsW is at (5,5) (top left) in the global window
        for (i = 1; i < typed.length()+1 && besti == -1; i++) {
          nextcharpos = x + 5 - textbox.textWidth(typed)/2 + textbox.textWidth(typed.substring(0,i));
          if (abs(mouseX - nextcharpos) > abs(mouseX - prevcharpos)) besti = i-1;
          prevcharpos = nextcharpos;
        }
        if (besti == -1) besti = typed.length();
        CursorCharIndex = besti;
        cursorbuffer = typed.substring(0,besti);
      }
    }
    else OnFocus = false;
  }

  void display() {
    textbox.beginDraw();
    TextOverlay();
    textbox.endDraw();
    menuL.image(border, x, y);
    menuL.image(textbox, x, y);
  }

  void TextOverlay() {
    float cpos; // cursor position (x)
    textbox.pushMatrix();
    textbox.smooth();
    buildtrickery();
    border.image(trickery, TWidth/2, THeight/2);
    textbox.image(trickery, textbox.width/2, textbox.height/2);
    textbox.fill(TColour);
    textbox.textFont(fontButton, Fontsize);
    textbox.textAlign(CENTER, CENTER);

    cpos = textbox.width/2 + textbox.textWidth(typed)/2 - (textbox.textWidth(typed) - textbox.textWidth(cursorbuffer));

    if (cpos > 0.9*textbox.width) textbox.translate(-(cpos - textbox.width + Margin*TWidth), 0);
    if (cpos < 0.1*textbox.width) textbox.translate(-(cpos - textbox.width + (1 - Margin)*TWidth), 0);

    textbox.text(typed, textbox.width/2, textbox.height/2);
    textbox.stroke(0);
    if(OnFocus) {
      textbox.line(cpos, 0.1*textbox.height, cpos, 0.9*textbox.height);
    }
    if(!OnFocus) {
      CursorCharIndex = typed.length()/2; 
      cursorbuffer = typed.substring(0,CursorCharIndex);
    }
    textbox.popMatrix();
  }

  String getText() {
    return typed;
  }

  void setText(String _typed) {
    typed = cursorbuffer = _typed;
     CursorCharIndex = typed.length();
  }
}

class FloatNumberBox extends Textbox {
  float Data; // the floating-point number entered
  float MaxData, MinData, Epslon; // working range
  int digsleft, digsright; // formatting options; Digits to the left of decimal separator (Integer) and Digits to the right (Fractionary)

  FloatNumberBox(int _x, int _y, int _TWidth, int _THeight, float _defaultvalue, int _Border, 
  float _MaxData, float _MinData, float _Epslon, int _digsleft, int _digsright, float _Margin, color _TColour) {
    super (_x,_y,_TWidth,_THeight,"",_Border,_Margin,_TColour); // won't compile using the usual way. This is workaround
    MaxData = (_MaxData > 9.9999999999999999e37) ? (9.9999999999999999e37) : (_MaxData); // Avoid Infinity
    Epslon = (_Epslon < 1e-44) ? (1e-44) : (_Epslon); // Tolerance around zero (anything within this tolerance will be replaced by strict zero)
    MinData = (_MinData < -9.9999999999999999e37) ? (-9.9999999999999999e37) : (_MinData); // Avoid -Infinity
    digsleft = _digsleft;
    digsright = _digsright;
    Data = _defaultvalue; 
    typed = cursorbuffer = (abs(Data) < 1.0e6) ? (nfp(FloatRound(Data, digsright), digsleft, digsright)) : ( (Data > 0) ? ("+" + str(Data)) : (str(Data)) ); 
    CursorCharIndex = typed.length();
    if (Data > MaxData) {
      typed = (abs(MaxData) < 1.0e6) ? (nfp(FloatRound(MaxData, digsright), digsleft, digsright)) : ( (MaxData > 0) ? ("+" + str(MaxData)) : (str(MaxData)) ); 
      CursorCharIndex = typed.length(); 
      Data = MaxData;
    }
    if (Data < MinData) {
      typed = (abs(MinData) < 1.0e6) ? (nfp(FloatRound(MinData, digsright), digsleft, digsright)) : ( (MinData > 0) ? ("+" + str(MinData)) : (str(MinData)) ); 
      CursorCharIndex = typed.length(); 
      Data = MinData;
    }
    if (abs(Data) <= Epslon || str(Data).equals("NaN")) {
      typed = cursorbuffer = nfp(0, digsleft, digsright); 
      CursorCharIndex = typed.length(); 
      Data = 0;
    }
  }

  float FloatRound(float number, float decimal) { // CAUTION: round() does not work with very big numbers (> 1e7). Probably stack overflow
    return (float)(round((number*pow(10, decimal))))/pow(10, decimal);
    
  }
  // http://processing.org/discourse/yabb2/YaBB.pl?num=1115073288
  void keyPressed() {
    if (OnFocus) KeyFilter();
  }

  void mouseClicked() {
    super.mouseClicked();
    if (!OnFocus) {
      typed = (abs(Data) < 1.0e6) ? (nfp(FloatRound(Data, digsright), digsleft, digsright)) : ( (Data > 0) ? ("+" + str(Data)) : (str(Data)) ); 
      CursorCharIndex = typed.length()/2; 
      cursorbuffer = typed.substring(0,CursorCharIndex);
    }
  }

  void KeyFilter() { 
    /* Filter non-number characters and error checking
     basic error checking
     
     only call KeyControl() if everything is OK
     then, convert the typed string into a floating-point number
     */
    int i, j, dotpos, exppos;
    boolean errorflag = false;
    float DataIncrement;
    String TMP;

    switch(key) {
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
    case '0': 
      if (typed.length() > CursorCharIndex &&
        (typed.charAt(CursorCharIndex) == '+' || typed.charAt(CursorCharIndex) == '-' )) {
        errorflag = true;
      }
      // Avoid inserting number before "-" or "+"
    case CODED:
       switch(keyCode) {
      case UP: // fix: if textwidth changes(+ and -), update cursorbuffer    ;;    fix: exponent tweaking
        // find nearest dot or comma before current character to see whether the cursor stands over a non-integer part of the number
        dotpos = -1;  // if dotpos remains -1, is integer
        exppos = typed.length(); // if exppos remains typed.length(), there is no 'e' (CursorCharIndex cannot be bigger than this)
        for (i = 0; i < typed.length(); i++) {
          if (typed.charAt(i) == '.') dotpos = i;
          if (typed.charAt(i) == 'e' || typed.charAt(i) == 'E') exppos = i;
        }
        if (CursorCharIndex <= exppos) {
          if (dotpos == -1 && CursorCharIndex != 0) {
            DataIncrement = pow(10,typed.length() - CursorCharIndex);
          } 
          else {
            DataIncrement = (CursorCharIndex <= dotpos) ? (pow(10, -(CursorCharIndex - dotpos) )) : (pow(10, -(CursorCharIndex - dotpos - 1) ));
          }
          Data += (Data + DataIncrement < MaxData) ? (DataIncrement) : (0);
          typed = (Data < 1.0e6) ? (nfp(FloatRound(Data, digsright), digsleft, digsright)) : ( (Data > 0) ? ("+" + str(Data)) : (str(Data)) ); 
          if (CursorCharIndex > typed.length()) CursorCharIndex = typed.length();
          cursorbuffer = typed.substring(0, CursorCharIndex);
        } 
        else {
          TMP = typed.substring(exppos+1, typed.length()); // get current exponent
          DataIncrement = pow(10, pow(10, TMP.length() +1 -(CursorCharIndex -exppos) ));
          if (TMP.charAt(0) == '-') { // negative exponent; rising will divide number
            Data /= DataIncrement;
          }
          else {
            Data *= DataIncrement;
          }
          typed = (Data < 1.0e6) ? (nfp(FloatRound(Data, digsright), digsleft, digsright)) : ( (Data > 0) ? ("+" + str(Data)) : (str(Data)) ); 
          if (CursorCharIndex > typed.length()) CursorCharIndex = typed.length();
          cursorbuffer = typed.substring(0, CursorCharIndex);
        }
        break;
      case DOWN:
        // find nearest dot or comma before current character to see whether the cursor stands over a non-integer part of the number
        dotpos = -1;  // if dotpos remains -1, is integer
        exppos = typed.length(); // if exppos remains typed.length(), there is no 'e' (CursorCharIndex cannot be bigger than this)
        for (i = 0; i < typed.length(); i++) {
          if (typed.charAt(i) == '.') dotpos = i;
          if (typed.charAt(i) == 'e' || typed.charAt(i) == 'E') exppos = i;
        }
        if (CursorCharIndex <= exppos) {
          if (dotpos == -1 && CursorCharIndex != 0) {
            DataIncrement = pow(10,typed.length() - CursorCharIndex);
          } 
          else {
            DataIncrement = (CursorCharIndex <= dotpos) ? (pow(10, -(CursorCharIndex - dotpos) )) : (pow(10, -(CursorCharIndex - dotpos - 1) ));
          }
          Data -= (Data - DataIncrement > MinData) ? (DataIncrement) : (0);
          typed = (Data < 1.0e6) ? (nfp(FloatRound(Data, digsright), digsleft, digsright)) : ( (Data > 0) ? ("+" + str(Data)) : (str(Data)) ); 
          if (CursorCharIndex > typed.length()) CursorCharIndex = typed.length();
          cursorbuffer = typed.substring(0, CursorCharIndex);
        } 
        else {
          TMP = typed.substring(exppos+1, typed.length()); // get current exponent
          DataIncrement = pow(10, pow(10, TMP.length() +1 -(CursorCharIndex -exppos) ));
          if (TMP.charAt(0) == '-') { // negative exponent; rising will multiply number
            Data *= DataIncrement;
          }
          else {
            Data /= DataIncrement;
          }
          typed = (Data < 1.0e6) ? (nfp(FloatRound(Data, digsright), digsleft, digsright)) : ( (Data > 0) ? ("+" + str(Data)) : (str(Data)) ); 
          if (CursorCharIndex > typed.length()) CursorCharIndex = typed.length();
          cursorbuffer = typed.substring(0, CursorCharIndex);
        }

        break;
      default: 
        if (!errorflag) KeyControl(); 
        break;
      }
      break;
    case DELETE:
    
    case BACKSPACE:
      KeyControl();
      if (typed.length() == 0) {
        typed = "0.0"; 
        Data = 0; 
        cursorbuffer = "0"; 
        CursorCharIndex = 1;
      } // Avoid emptyness causing NaN
      if (CursorCharIndex > 1 && 
        (typed.charAt(CursorCharIndex - 1) == 'e' || typed.charAt(CursorCharIndex - 1) == 'E') && 
        typed.length() == CursorCharIndex) {
        typed += '0';
      } // Avoid 'e' and 'E' as last char
      if (CursorCharIndex > 1 && 
        CursorCharIndex == typed.length() && 
        (typed.charAt(CursorCharIndex - 1) == '+' || typed.charAt(CursorCharIndex - 1) == '-')) {
        typed += '0';
      } 
      // avoid '-' and '+' as last char
      if (typed.length() > CursorCharIndex &&
        (typed.charAt(0) == '+' || typed.charAt(0) == '-') &&
        (typed.charAt(CursorCharIndex) == '+' || typed.charAt(CursorCharIndex) == '-')) 
      {
        typed = typed.substring(0,CursorCharIndex) + typed.substring(CursorCharIndex+1,typed.length());
      }
      // Avoid "++" , "+-", "-+", and "--"; Avoid "+e", "-e", "+E", and "-E". Apagando e tinha notacao cientifica E sinal no numero
      // apagando um 'e', apaga tb o sinal q vem depois
      if (typed.length() > CursorCharIndex &&
        ((typed.charAt(0) == 'e' || typed.charAt(0) == 'E') || 
        ((typed.charAt(0) == '+' || typed.charAt(0) == '-') && 
        (typed.charAt(1) == 'e' || typed.charAt(1) == 'E' ) ) )) {
        typed = typed.substring(0,CursorCharIndex) + '1' + typed.substring(CursorCharIndex,typed.length()); 
        CursorCharIndex++; 
        cursorbuffer = typed.substring(0, CursorCharIndex);
      }
      // Avoid "e" or "E" as first char or as second char if first is "-" or "+"
      break;
    case '.':
      for(i = 0; i < typed.length(); i++) {
        if (typed.charAt(i) == '.') errorflag = true;
      }
      if (!errorflag) {
        KeyControl(); 
        Data = float(typed);
      }
      break;
    case '+': // notacao cientifica: 1.34565e+10 e sinal: +1.23456e+10. Soh pode colocar fora da posicao de sinal se tiver o 'e'
      if (((typed.charAt(0) == '-' || typed.charAt(0) == '+') && 
        (CursorCharIndex == 1 || CursorCharIndex == 0)) ||
        ( CursorCharIndex > 0 &&
        ((typed.charAt(CursorCharIndex-1) != 'e' && typed.charAt(CursorCharIndex-1) != 'E')) ) ) {
        errorflag = true;
      }
      for (i = 1; i < typed.length(); i++) {
        if ((typed.charAt(i) == '+' || typed.charAt(i) == '-') &&
          CursorCharIndex != 0) {
          errorflag = true;
        }
      }
      if (!errorflag) {
        KeyControl();
      }
      break;
    case '-': // notacao cientifica: 1.2345e-10 e sinal: -12345e+10. Soh pode colocar fora da posicao de sinal se tiver o 'e'
      if (((typed.charAt(0) == '-' || typed.charAt(0) == '+') && 
        (CursorCharIndex == 1 || CursorCharIndex == 0)) ||
        ( CursorCharIndex > 0 &&
        ((typed.charAt(CursorCharIndex-1) != 'e' && typed.charAt(CursorCharIndex-1) != 'E')) ) ) {
        errorflag = true;
      } 
      // n pode colocar ++ ou +- ou -- ou -+ como sinal E nem colocar no meio se n tiver o 'e'
      for (i = 1; i < typed.length(); i++) {
        if ((typed.charAt(i) == '-' || typed.charAt(i) == '+') &&
          CursorCharIndex != 0) {
          errorflag = true;
        }
      }
      if (!errorflag) {
        KeyControl();
      }
      break;
    case 'e':
    case 'E':
      for (i = 1; i < typed.length(); i++) {
        if (typed.charAt(i) == 'e' || typed.charAt(i) == 'E') errorflag = true;
        if (CursorCharIndex < typed.length()) {
          for (j = CursorCharIndex; j < typed.length(); j++) {
            if (typed.charAt(j) == '.' || typed.charAt(j) == ',') errorflag = true;
          }
        }
      }
      if (!errorflag) {
        KeyControl();
        if (CursorCharIndex == typed.length()) typed += '0'; // avoid NaN errors
        if (CursorCharIndex == 1 || 
          (typed.charAt(CursorCharIndex - 2) == '+'
          || typed.charAt(CursorCharIndex - 2) == '-' )) {
          typed = typed.substring(0,CursorCharIndex-1) + '1' + typed.substring(CursorCharIndex-1,typed.length()); 
          CursorCharIndex++; 
          cursorbuffer = typed.substring(0, CursorCharIndex);
        }
      }
      break;
    case RETURN:
    case ENTER: 
      OnFocus = false; 
      
      break;
    }
    Data = float(typed);
    if (Data > MaxData) {
      typed = cursorbuffer = (abs(MaxData) < 1.0e6) ? (nfp(FloatRound(MaxData, digsright), digsleft, digsright)) : ( (MaxData > 0) ? ("+" + str(MaxData)) : (str(MaxData)) ); 
      CursorCharIndex = typed.length(); 
      Data = MaxData;
    }
    if (Data < MinData) {
      typed = cursorbuffer = (abs(MinData) < 1.0e6) ? (nfp(FloatRound(MinData, digsright), digsleft, digsright)) : ( (MinData > 0) ? ("+" + str(MinData)) : (str(MinData)) ); 
      CursorCharIndex = typed.length(); 
      Data = MinData;
    }
    if (!OnFocus) {
      if (abs(Data) <= Epslon || str(Data).equals("NaN")) {
        typed = cursorbuffer = str(0); 
        CursorCharIndex = typed.length(); 
        Data = 0;
      }
      // round whatever is smaller than Epslon and if, with all the exhaustive error checking, the user is STILL able to NaN things up, fix
      
        else {
        typed = (abs(Data) < 1.0e6) ? (nfp(FloatRound(Data, digsright), digsleft, digsright)) : ( (Data > 0) ? ("+" + str(Data)) : (str(Data)) ); 
        CursorCharIndex = typed.length()/2; 
        cursorbuffer = typed.substring(0,CursorCharIndex);
      }
    }
  }

  float getData() {
    return Data;
  }

  void setData(float _Data) {
    Data = _Data;
    typed = (abs(Data) < 1.0e6) ? (nfp(FloatRound(Data, digsright), digsleft, digsright)) : ( (Data > 0) ? ("+" + str(Data)) : (str(Data)) ); 
    CursorCharIndex = typed.length()/2; 
    cursorbuffer = typed.substring(0,CursorCharIndex);
  }
}
